from datetime import datetime, UTC

from bottle import route, redirect, request

from matemat.db import MatematDatabase
from matemat.exceptions import AuthenticationError
from matemat.webserver import template, session
from matemat.webserver.template import Notification
from matemat.webserver.config import get_app_config, get_stock_provider
from matemat.util.currency_format import format_chf


@route('/')
def main_page():
    """
    The main page, showing either the user list (if no user is logged in) or the product list (if a user is logged in).
    """
    config = get_app_config()
    session_id: str = session.start()
    now = str(int(datetime.now(UTC).timestamp()))
    with MatematDatabase(config['DatabaseFile']) as db:
        # Fetch the list of products to display
        products = db.list_products()
        buyproduct = None

        if request.params.barcode:
            try:
                buyproduct = db.get_product_by_barcode(request.params.barcode)
            except ValueError:
                if not session.has(session_id, 'authenticated_user'):
                    try:
                        user, token = db.tokenlogin(str(request.params.barcode))
                        # Set the user ID session variable
                        session.put(session_id, 'authenticated_user', user.id)
                        # Set the authlevel session variable (0 = none, 1 = token, 2 = touchkey, 3 = password)
                        session.put(session_id, 'authentication_level', 1)
                        redirect('/')
                    except AuthenticationError:
                        # Redirect to main page on token login error
                        pass
                Notification.error(f'Barcode {request.params.barcode} is not associated with any product.', decay=True)
                redirect('/')

        # Check whether a user is logged in
        if session.has(session_id, 'authenticated_user'):
            # Fetch the user id and authentication level (touchkey vs password) from the session storage
            uid: int = session.get(session_id, 'authenticated_user')
            authlevel: int = session.get(session_id, 'authentication_level')
            # If an barcode was scanned, directly trigger the purchase
            if buyproduct:
                redirect(f'/buy?pid={buyproduct.id}')
            # Fetch the user object from the database (for name display, price calculation and admin check)
            users = db.list_users()
            user = db.get_user(uid)
            if 'logout_after_purchase' in request.params:
                db.change_user(user, agent=user, logout_after_purchase=request.params.logout_after_purchase != '0')
            # Prepare a response with a jinja2 template
            return template.render('productlist.html',
                                   authuser=user, users=users, products=products, authlevel=authlevel,
                                   stock=get_stock_provider(), setupname=config['InstanceName'], now=now)
        else:
            # If there are no admin users registered, jump to the admin creation procedure
            if not db.has_admin_users():
                redirect('/userbootstrap')
            if buyproduct:
                Notification.success(
                    f'Login will purchase <strong>{buyproduct.name}</strong>. ' +
                    'Click <a class="alert-link" href="/">here</a> to abort.')
            # If no user is logged in, fetch the list of users and render the userlist template
            users = db.list_users(with_touchkey=True)
            return template.render('userlist.html',
                                   users=users, setupname=config['InstanceName'], now=now,
                                   signup=(config.get('SignupEnabled', '0') == '1'),
                                   buyproduct=buyproduct)
