import os
from datetime import datetime, UTC
from typing import Dict, Optional

from bottle import get, post, redirect, abort, request, FormsDict

from matemat.db import MatematDatabase
from matemat.db.primitives import User, ReceiptPreference
from matemat.exceptions import DatabaseConsistencyError
from matemat.util.currency_format import parse_chf
from matemat.util.thumbnails import upload_thumbnail
from matemat.webserver import template, session
from matemat.webserver.config import get_app_config
from matemat.webserver.template import Notification


@get('/moduser')
@post('/moduser')
def moduser():
    """
    The user modification page available from the admin panel.
    """
    config = get_app_config()
    session_id: str = session.start()
    # If no user is logged in, redirect to the login page
    if not session.has(session_id, 'authentication_level') or not session.has(session_id, 'authenticated_user'):
        redirect('/login')
    authlevel: int = session.get(session_id, 'authentication_level')
    auth_uid: int = session.get(session_id, 'authenticated_user')
    # Show a 403 Forbidden error page if no user is logged in (0) or a user logged in via token (1) / touchkey (2)
    if authlevel < 3:
        abort(403)

    # Connect to the database
    with MatematDatabase(config['DatabaseFile']) as db:
        # Fetch the authenticated user
        authuser = db.get_user(auth_uid)
        if not authuser.is_admin:
            # Show a 403 Forbidden error page if the user is not an admin
            abort(403)
        if 'userid' not in request.params:
            # Show a 400 Bad Request error page if no users to edit was specified
            # (should never happen during normal operation)
            abort(400, '"userid" argument missing')

        # Fetch the user to modify from the database
        moduser_id = int(str(request.params.userid))
        user = db.get_user(moduser_id)

        # If the request contains a "change" parameter, delegate the change handling to the function below
        if 'change' in request.params:
            handle_change(request.params, request.files, user, authuser, db)
            # If the user was deleted, redirect back to the admin page, as there is nothing to edit any more
            if str(request.params.change) == 'del':
                redirect('/admin')

        # Render the "Modify User" page
        tokens = db.list_tokens(moduser_id)
        now = str(int(datetime.now(UTC).timestamp()))
        return template.render('moduser.html',
                               authuser=authuser, user=user, authlevel=authlevel, now=now,
                               receipt_preference_class=ReceiptPreference, tokens=tokens,
                               setupname=config['InstanceName'], config_smtp_enabled=config['SmtpSendReceipts'])


def handle_change(args: FormsDict, files: FormsDict, user: User, authuser: User, db: MatematDatabase) \
        -> None:
    """
    Write the changes requested by an admin to the database.

    :param args: The RequestArguments object passed to the pagelet.
    :param user: The user to edit.
    :param authuser: The user performing the modification.
    :param db: The database facade where changes are written to.
    """
    config = get_app_config()
    # Read the type of change requested by the admin, then switch over it
    change = str(args.change)

    # Admin requested deletion of the user
    if change == 'del':
        # Delete the user from the database
        db.delete_user(user)
        # Delete the user's avatar, if it exists
        try:
            abspath: str = os.path.join(os.path.abspath(config['UploadDir']), 'thumbnails/users/')
            os.remove(os.path.join(abspath, f'{user.id}.png'))
        except FileNotFoundError:
            pass

    elif change == 'deltoken':
        try:
            tokid = id(str(request.params.token))
            token = db.get_token(tokid)
        except Exception as e:
            Notification.error('Token not found', decay=True)
            return
        try:
            db.delete_token(token)
        except DatabaseConsistencyError:
            Notification.error(f'Failed to delete token {token.name}', decay=True)
        Notification.success(f'Token {token.name} removed', decay=True)
        return

    # Admin requested update of the user's details
    elif change == 'update':
        # Only write a change if all properties of the user are present in the request arguments
        if 'username' not in args or \
                'email' not in args or \
                'password' not in args or \
                'balance' not in args or \
                'receipt_pref' not in args:
            return
        # Read the properties from the request arguments
        username = str(args.username)
        email = str(args.email)
        try:
            receipt_pref = ReceiptPreference(int(str(args.receipt_pref)))
        except ValueError:
            return
        password = str(args.password)
        balance = parse_chf(str(args.balance))
        balance_reason: Optional[str] = str(args.reason)
        if balance_reason == '':
            balance_reason = None
        is_member = 'ismember' in args
        is_admin = 'isadmin' in args
        logout_after_purchase = 'logout_after_purchase' in args
        # An empty e-mail field should be interpreted as NULL
        if len(email) == 0:
            email = None
        # Attempt to write the changes to the database
        try:
            # If a password was entered, replace the password in the database
            if len(password) > 0:
                db.change_password(user, '', password, verify_password=False)
            # Write the user detail changes
            db.change_user(user, agent=authuser, name=username, email=email, is_member=is_member, is_admin=is_admin,
                           balance=balance, balance_reason=balance_reason, receipt_pref=receipt_pref,
                           logout_after_purchase=logout_after_purchase)
        except DatabaseConsistencyError:
            return
        # If a new avatar was uploaded, process it
        if 'avatar' in files:
            # Read the raw image data from the request
            image = files.avatar.file.read()
            filename: str = os.path.join(os.path.abspath(config['UploadDir']), f'thumbnails/users/{user.id}.png')
            try:
                upload_thumbnail(image, filename)
            except Exception as e:
                Notification.error(str(e), decay=True)
