
import os

from bottle import get, post, redirect, abort, request, FormsDict
import netaddr

from matemat.db import MatematDatabase
from matemat.db.primitives import User
from matemat.util.thumbnails import upload_thumbnail
from matemat.webserver import template, session
from matemat.webserver.config import get_app_config


def signup_user(args: FormsDict, files: FormsDict, db: MatematDatabase) -> User:
    config = get_app_config()
    if 'username' not in args or 'password' not in args or 'password2' not in args:
        raise ValueError('Required fields missing')
    username = str(args.username)
    password = str(args.password)
    password2 = str(args.password2)
    if password != password2:
        raise ValueError('Passwords do not match')
    email = None
    if 'email' in args and len(args.email) > 0:
        email = str(args.email)

    new_user = db.create_user(username, password, email=email, admin=False, member=False)
    # If a touchkey was provided, set it
    if 'touchkey' in args and len(args.touchkey) > 0:
        touchkey = str(args.touchkey)
        db.change_touchkey(new_user, password, touchkey, verify_password=False)
    # Finally, set the avatar, if provided
    image = files.avatar.file.read() if 'avatar' in files else None
    filename: str = os.path.join(os.path.abspath(config['UploadDir']), f'thumbnails/users/{new_user.id}.png')
    upload_thumbnail(image, filename)

    return new_user


@get('/signup')
@post('/signup')
def signup():
    """
    The password login mechanism.  If called via GET, render the UI template; if called via POST, attempt to log in with
    the provided credentials (username and passsword).
    """
    config = get_app_config()
    session_id: str = session.start()
    # If signup is not enabled, redirect to the main page
    if config.get('SignupEnabled', '0') != '1':
        redirect('/')
    # If a user is already logged in, simply redirect to the main page, showing the product list
    if session.has(session_id, 'authenticated_user'):
        redirect('/')
    # If requested via HTTP POST, read the request arguments and attempt to create a new user
    if request.method == 'POST':
        # Connect to the database and create the user
        with MatematDatabase(config['DatabaseFile']) as db:
            try:
                user = signup_user(request.params, request.files, db)
            except Exception as e:
                Notification.error(str(e), decay=True)
                redirect('/signup')
            # Set the user ID session variable
            session.put(session_id, 'authenticated_user', user.id)
            # Set the authlevel session variable (0 = none, 1 = token, 2 = touchkey, 3 = password)
            session.put(session_id, 'authentication_level', 3)
            # Redirect to the main page, showing the product list
            redirect('/')
    elif request.method != 'GET':
        abort(405, 'Method not allowed')

    acl_addrs = [s.strip() for s in config.get('SignupKioskMode', '').split(',') if s.strip()]
    acl = netaddr.IPSet(acl_addrs)
    if request.remote_addr in acl:
        return template.render('signup_kiosk.html',
                               signup=(config.get('SignupEnabled', '0') == '1'),
                               zip=zip,
                               setupname=config['InstanceName'])
    return template.render('signup.html',
                           signup=(config.get('SignupEnabled', '0') == '1'),
                           setupname=config['InstanceName'])
